from django.db import models


class AcmeChallenge(models.Model):
    """
    Store Let's Encrypt .well-known/acme-challenge data
    """
    challenge = models.CharField(unique=True, max_length=255, help_text='The identifier for this challenge')
    response = models.CharField(max_length=255, help_text='The response expected for this challenge')


    class Meta:
        verbose_name = 'ACME Challenge'
        verbose_name_plural = 'ACME Challenges'

    def __str__(self):
        return self.challenge



from acme import challenges
from asymmetric_jwt_auth import generate_key_pair
from certbot import errors
from certbot.tests import acme_util
from certbot.tests import util as test_util
import tempfile
import unittest
import six
import mock
import requests_mock
import shutil
import os.path
import re


class AuthenticatorTest(unittest.TestCase):
    def setUp(self):
        self.http_achall = acme_util.HTTP01_A
        self.achalls = [self.http_achall]
        self.config = mock.MagicMock(
            http01_port=0, django_username=None, django_key_directory=None,
            django_public_ip_logging_ok=False, noninteractive_mode=False)

        self.temp_dir = tempfile.mkdtemp()

        from certbot_django.coordinator.authenticator import Authenticator
        self.auth = Authenticator(self.config, name='django')


    def tearDown(self):
        shutil.rmtree(self.temp_dir)


    def test_bad_key_dir(self):
        self.config.django_key_directory = '/this/dir/doest/exist/'
        self.config.django_public_ip_logging_ok = True
        self.assertRaises(errors.PluginError, self.auth.perform, self.achalls)


    def test_more_info(self):
        self.assertTrue(isinstance(self.auth.more_info(), six.string_types))


    def test_get_chall_pref(self):
        self.assertEqual(self.auth.get_chall_pref('example.org'), [challenges.HTTP01])


    @test_util.patch_get_utility()
    def test_ip_logging_not_ok(self, mock_get_utility):
        mock_get_utility().yesno.return_value = False
        self.assertRaises(errors.PluginError, self.auth.perform, [])


    @test_util.patch_get_utility()
    def test_ip_logging_ok(self, mock_get_utility):
        mock_get_utility().yesno.return_value = True
        self.auth.perform([])
        self.assertTrue(self.config.django_public_ip_logging_ok)


    @test_util.patch_get_utility()
    def test_perform_generate_key(self, mock_get_utility):
        self.config.django_public_ip_logging_ok = True
        self.config.django_username = 'certbot'
        self.config.django_key_directory = self.temp_dir

        with requests_mock.mock() as m:
            m.post('http://example.com/.well-known/challenges/', json={}, status_code=201)
            actual = self.auth.perform(self.achalls)
            expected = [achall.response(achall.account_key) for achall in self.achalls]
            self.assertEqual(actual, expected)
            self.assertEqual(m.call_count, 1)

        for i, (args, kwargs) in enumerate(mock_get_utility().notification.call_args_list):
            self.assertTrue("new key has been generated and saved" in args[0])
            self.assertTrue("user 'certbot'" in args[0])
            self.assertFalse(kwargs['wrap'])


    @test_util.patch_get_utility()
    def test_perform_existing_key(self, mock_get_utility):
        self.config.django_public_ip_logging_ok = True
        self.config.django_username = 'certbot'
        self.config.django_key_directory = self.temp_dir

        priv, pub = generate_key_pair()
        keypath = os.path.join(self.temp_dir, 'certbot_django_id_rsa_examplecom')
        with open(keypath, 'w') as keyfile:
            keyfile.write(priv)

        with requests_mock.mock() as m:
            m.post('http://example.com/.well-known/challenges/', json={}, status_code=201)
            actual = self.auth.perform(self.achalls)
            expected = [achall.response(achall.account_key) for achall in self.achalls]
            self.assertEqual(actual, expected)
            self.assertEqual(m.call_count, 1)

        for i, (args, kwargs) in enumerate(mock_get_utility().notification.call_args_list):
            self.assertEqual("" in args[0])
            self.assertFalse(kwargs['wrap'])


    @test_util.patch_get_utility()
    def test_perform_bad_auth(self, mock_get_utility):
        self.config.django_public_ip_logging_ok = True
        self.config.django_username = 'certbot'
        self.config.django_key_directory = self.temp_dir

        priv, pub = generate_key_pair()
        keypath = os.path.join(self.temp_dir, 'certbot_django_id_rsa_examplecom')
        with open(keypath, 'w') as keyfile:
            keyfile.write(priv)

        with requests_mock.mock() as m:
            m.post('http://example.com/.well-known/challenges/', json={}, status_code=403)
            self.assertRaises(errors.PluginError, self.auth.perform, self.achalls)


    def test_cleanup(self):
        self.config.django_public_ip_logging_ok = True
        self.config.django_username = 'certbot'
        self.config.django_key_directory = self.temp_dir

        priv, pub = generate_key_pair()
        keypath = os.path.join(self.temp_dir, 'certbot_django_id_rsa_examplecom')
        with open(keypath, 'w') as keyfile:
            keyfile.write(priv)

        with requests_mock.mock() as m:
            m.post('http://example.com/.well-known/challenges/', json={}, status_code=201)
            m.delete(re.compile('example\.com/\.well-known/challenges/[A-Za-z0-9]+/$'), json={}, status_code=204)
            self.auth.perform(self.achalls)
            self.auth.cleanup(self.achalls)
            self.assertEqual(m.call_count, 2)



from rest_framework import serializers
from rest_framework.reverse import reverse
from .models import AcmeChallenge


class AcmeChallengeSerializer(serializers.ModelSerializer):
    acme_url = serializers.SerializerMethodField()

    class Meta:
        model = AcmeChallenge
        fields = ('challenge', 'response', 'acme_url')

    def get_acme_url(self, obj):
        return reverse(viewname='acmechallenge-response', args=(obj.challenge, ), request=self.context['request'])



from django.contrib import admin
from django.utils.html import format_html
from django.core.urlresolvers import reverse, NoReverseMatch
from .models import AcmeChallenge


class AcmeChallengeAdmin(admin.ModelAdmin):
    def format_acme_url(self, acme_object):
        try:
            object_url = reverse(viewname='acmechallenge-response', args=(acme_object.challenge, ))
            return format_html("<a href='{}'>ACME Challenge Link</a>", object_url)
        except NoReverseMatch:
            return '-'
    format_acme_url.short_description = 'Link'

    fieldsets = [
        ('ACME Request', {
            'fields': [
                'challenge',
                'response',
            ],
        }),
        ('Metadata', {
            'fields': [
                'id',
                'format_acme_url',
            ],
        }),
    ]

    list_display = ['challenge', 'format_acme_url']
    ordering = ['challenge']
    readonly_fields = ['id', 'format_acme_url']
    search_fields = ['challenge', 'response']


admin.site.register(AcmeChallenge, AcmeChallengeAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-02-21 20:15
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='AcmeChallenge',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('challenge', models.CharField(help_text='The identifier for this challenge', max_length=255, unique=True)),
                ('response', models.CharField(help_text='The response expected for this challenge', max_length=255)),
            ],
            options={
                'verbose_name': 'ACME Challenge',
                'verbose_name_plural': 'ACME Challenges',
            },
        ),
    ]



from django.shortcuts import get_object_or_404, render
from rest_framework import viewsets, permissions
from .models import AcmeChallenge
from .serializers import AcmeChallengeSerializer


class AcmeChallengeViewSet(viewsets.ModelViewSet):
    queryset = AcmeChallenge.objects.all()
    serializer_class = AcmeChallengeSerializer
    lookup_field = 'challenge'
    permission_classes = (permissions.IsAdminUser, permissions.DjangoModelPermissions)


def detail(request, acme_data):
    acme_challenge = get_object_or_404(AcmeChallenge, challenge=acme_data)
    context = {
        'response': acme_challenge.response
    }
    return render(request, 'certbot_django/detail.html', context)



from django.conf.urls import include, url
from rest_framework.routers import DefaultRouter
from . import views


router = DefaultRouter()
router.register(r'challenges', views.AcmeChallengeViewSet)


urlpatterns = [
    url(r'^acme-challenge/(?P<acme_data>.+)$', views.detail, name='acmechallenge-response'),
    url(r'^', include(router.urls)),
]



from django.test import TestCase
from ..models import AcmeChallenge


class TestAcmeChallenge(TestCase):
    def test_challenge(self):
        """
        Test the challenge data in the model
        """
        challenge = 'challenge'
        response = ''
        acme_object = AcmeChallenge(challenge=challenge, response=response)
        self.assertEqual(acme_object.challenge, challenge)


    def test_response(self):
        """
        Test the response data in the model
        """
        challenge = ''
        response = 'challenge.response'
        acme_object = AcmeChallenge(challenge=challenge, response=response)
        self.assertEqual(acme_object.response, response)


    def test_str(self):
        """
        Test the __str__ representation
        """
        challenge = 'challenge'
        response = 'challenge.response'
        acme_object = AcmeChallenge(challenge=challenge, response=response)
        self.assertEqual(str(acme_object), challenge)



from asymmetric_jwt_auth.models import PublicKey
from asymmetric_jwt_auth import generate_key_pair, create_auth_header
from django.contrib.auth.models import User, Permission
from django.core.urlresolvers import reverse
from django.test import TestCase
from rest_framework import status
from ..models import AcmeChallenge


class TestAcmeChallengeViews(TestCase):
    def setUp(self):
        self.expected_challenge = 'challenge_view_test'
        self.expected_response = 'challenge_view_test_response'
        self.expected_response_bytes = b'challenge_view_test_response\n'
        self.expected_response_decode = 'challenge_view_test_response\n'
        self.test_challenge = AcmeChallenge.objects.create(challenge=self.expected_challenge, response=self.expected_response)


    def test_detail(self):
        """
        When given a valid request challenge, make sure the response is returned
        """
        url = reverse('acmechallenge-response', args=(self.expected_challenge, ))
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.content, self.expected_response_bytes)
        self.assertEqual(response.content.decode(response.charset), self.expected_response_decode)


    def test_detail_404(self):
        """
        When given a bad request challenge, make sure a 404 is returned
        """
        url = reverse('acmechallenge-response', args=('fake', ))
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)



class TestAcmeChallengeAPI(TestCase):
    def setUp(self):
        self.perm_add = Permission.objects.get(codename='add_acmechallenge')
        self.perm_change = Permission.objects.get(codename='change_acmechallenge')
        self.perm_delete = Permission.objects.get(codename='delete_acmechallenge')

        self.user_joe = User.objects.create_user(username='joe')
        self.user_joe.is_staff = False
        self.user_joe.save()

        self.user_certbot = User.objects.create_user(username='certbot')
        self.user_certbot.is_staff = True
        self.user_certbot.user_permissions = [
            self.perm_add,
            self.perm_change,
            self.perm_delete,
        ]
        self.user_certbot.save()

        _priv, _pub = generate_key_pair()
        self.priv_key_joe = _priv
        PublicKey.objects.create(key=_pub, comment='Test Key', user=self.user_joe)

        _priv, _pub = generate_key_pair()
        self.priv_key_certbot = _priv
        PublicKey.objects.create(key=_pub, comment='Test Key', user=self.user_certbot)


    def test_add_challenge(self):
        matrix = (
            ('', '', status.HTTP_403_FORBIDDEN),
            ('joe', self.priv_key_joe, status.HTTP_403_FORBIDDEN),
            ('certbot', self.priv_key_joe, status.HTTP_403_FORBIDDEN),
            ('certbot', self.priv_key_certbot, status.HTTP_201_CREATED),
        )

        url = reverse('acmechallenge-list')

        data = {
            'challenge': 'foo',
            'response': 'bar',
        }

        for username, private_key, expected_status in matrix:
            headers = {}
            if username:
                headers['HTTP_AUTHORIZATION'] = create_auth_header(username=username, key=private_key)
            response = self.client.post(url, data=data, **headers)
            self.assertEquals(response.status_code, expected_status)


    def test_remove_challenge(self):
        matrix = (
            ('', '', status.HTTP_403_FORBIDDEN),
            ('joe', self.priv_key_joe, status.HTTP_403_FORBIDDEN),
            ('certbot', self.priv_key_joe, status.HTTP_403_FORBIDDEN),
            ('certbot', self.priv_key_certbot, status.HTTP_204_NO_CONTENT),
        )

        for username, private_key, expected_status in matrix:
            AcmeChallenge.objects.get_or_create(challenge='foo', response='bar')
            url = reverse('acmechallenge-detail', args=('foo', ))
            headers = {}
            if username:
                headers['HTTP_AUTHORIZATION'] = create_auth_header(username=username, key=private_key)
            response = self.client.delete(url, **headers)
            self.assertEquals(response.status_code, expected_status)


    def test_view_challenge(self):
        matrix = (
            ('', '', status.HTTP_403_FORBIDDEN),
            ('joe', self.priv_key_joe, status.HTTP_403_FORBIDDEN),
            ('certbot', self.priv_key_joe, status.HTTP_403_FORBIDDEN),
            ('certbot', self.priv_key_certbot, status.HTTP_200_OK),
        )

        for username, private_key, expected_status in matrix:
            AcmeChallenge.objects.get_or_create(challenge='foo', response='bar')
            url = reverse('acmechallenge-detail', args=('foo', ))
            headers = {}
            if username:
                headers['HTTP_AUTHORIZATION'] = create_auth_header(username=username, key=private_key)
            response = self.client.get(url, **headers)
            self.assertEquals(response.status_code, expected_status)
            if expected_status == status.HTTP_200_OK:
                self.assertEquals(response.data['challenge'], 'foo')
                self.assertEquals(response.data['response'], 'bar')



#!/usr/bin/env python
from setuptools import setup, find_packages, Distribution
import codecs
import os.path

# Make sure versiontag exists before going any further
Distribution().fetch_build_eggs('versiontag>=1.2.0')

from versiontag import get_version, cache_git_tag  # NOQA


packages = find_packages('src')

install_requires = [
    'asymmetric_jwt_auth>=0.3.0',
    'Django>=1.8.17',
    'djangorestframework>=3.4.7',
    'requests>=2.13.0'
]

extras_require = {
    'coordinator': [
        'acme>=0.1.1',
        'certbot>=0.9.3',
        'zope.interface',
    ],
    'development': [
        'flake8>=3.2.1',
        'requests-mock>=1.3.0',
        'sphinx>=1.5.2',
        'versiontag>=1.2.0',
    ],
}

entry_points = {
    'certbot.plugins': [
        'auth = certbot_django.coordinator.authenticator:Authenticator',
    ],
}


def fpath(name):
    return os.path.join(os.path.dirname(__file__), name)


def read(fname):
    return codecs.open(fpath(fname), encoding='utf-8').read()


cache_git_tag()

setup(
    name='certbot-django',
    description="Full-stack Django / LetsEncrypt / Certbot Integration",
    version=get_version(pypi=True),
    long_description=open('README.rst').read(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Plugins',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Security',
        'Topic :: System :: Installation/Setup',
        'Topic :: System :: Networking',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',
    ],
    author="Craig Weber",
    author_email='crgwbr@gmail.com',
    url='https://gitlab.com/thelabnyc/certbot-django',
    license='ISC',
    package_dir={'': 'src'},
    packages=packages,
    include_package_data=True,
    install_requires=install_requires,
    extras_require=extras_require,
    entry_points=entry_points,
)



from acme import challenges
from asymmetric_jwt_auth import create_auth_header, generate_key_pair
from certbot import interfaces, errors
from certbot.plugins import common
from certbot.display import util as display_util
import requests
import os.path
import os
import zope.component
import zope.interface
import logging

logger = logging.getLogger(__name__)


def _test_key_dir_read_write(key_dir):
    test_file_path = os.path.join(key_dir, 'certbot-test-file.txt')

    try:
        with open(test_file_path, 'w') as testfile:
            testfile.write('test')
    except (IOError, OSError):
        raise errors.PluginError('Could not write file to %s' % key_dir)

    try:
        with open(test_file_path, 'r') as testfile:
            testfile.readlines()
    except (IOError, OSError):
        raise errors.PluginError('Could not read file from %s' % key_dir)

    try:
        os.unlink(test_file_path)
    except (IOError, OSError):
        raise errors.PluginError('Could not remove file from %s' % key_dir)


def _validate_key_dir(key_dir):
    key_dir = os.path.expanduser(key_dir)
    if not os.path.isdir(key_dir):
        raise errors.PluginError(key_dir + " does not exist or is not a directory")
    key_dir = os.path.abspath(key_dir)
    _test_key_dir_read_write(key_dir)
    return key_dir


def _validate_username(username):
    if len(username) <= 0:
        raise errors.PluginError("Username can not be blank")
    return username


@zope.interface.implementer(interfaces.IAuthenticator)
@zope.interface.provider(interfaces.IPluginFactory)
class Authenticator(common.Plugin):
    """Django authenticator

    This plugin allows the user to perform the domain validation by
    using a remote-Django application that has certbot_django.server
    installed.
    """

    description = 'Django Server Authenticator'
    long_description = (
        'Authenticate by connecting to a Django application that has'
        'certbot_django.server installed.'
    )


    @classmethod
    def add_parser_arguments(cls, add):
        add('username',
            help='Username for authenticating with Django server.')
        add('key-directory',
            help='Directory to store generated public / private keys or to read previously generated keys from.')
        add('public-ip-logging-ok', action='store_true',
            help='Automatically allows public IP logging (default: Ask)')


    def prepare(self):
        pass


    def more_info(self):
        return (
            'Authenticate by connecting to a Django application that has'
            'certbot_django.server installed.'
        )


    def get_chall_pref(self, domain):
        """
        Only HTTP challenges are supported currently.
        """
        return [challenges.HTTP01]


    def perform(self, achalls):
        self._verify_ip_logging_ok()
        responses = []
        for achall in achalls:
            self._perform_achall(achall)
            responses.append(achall.response(achall.account_key))
        return responses


    def cleanup(self, achalls):
        for achall in achalls:
            domain = achall.domain
            challenge = achall.chall.encode('token')
            self._remove_challenge_from_server(domain, challenge)


    def _verify_ip_logging_ok(self):
        if not self.conf('public-ip-logging-ok'):
            cli_flag = '--{0}'.format(self.option_name('public-ip-logging-ok'))
            msg = ('NOTE: The IP of this machine will be publicly logged as '
                   "having requested this certificate. If you're running "
                   'certbot in manual mode on a machine that is not your '
                   "server, please ensure you're okay with that.\n\n"
                   'Are you OK with your IP being logged?')
            display = zope.component.getUtility(interfaces.IDisplay)
            if display.yesno(msg, cli_flag=cli_flag, force_interactive=True):
                setattr(self.config, self.dest('public-ip-logging-ok'), True)
            else:
                raise errors.PluginError('Must agree to IP logging to proceed')


    def _perform_achall(self, achall):
        domain = achall.domain
        challenge = achall.chall.encode('token')
        response = achall.validation(achall.account_key)
        self._add_challenge_to_server(domain, challenge, response)


    def _add_challenge_to_server(self, domain, challenge, response):
        url = 'http://{}/.well-known/challenges/'.format(domain)
        headers = self._get_headers(domain)
        data = {
            'challenge': challenge,
            'response': response,
        }
        try:
            logger.info("Attempting to add ACMEChallenge to server: %s" % challenge)
            resp = requests.post(url, headers=headers, data=data)
            resp.raise_for_status()
            logger.info("Successfully added ACMEChallenge to server: %s" % challenge)
        except requests.RequestException as e:
            raise errors.PluginError('Encountered error when adding challenge to Django server: {}'.format(e))


    def _remove_challenge_from_server(self, domain, challenge):
        url = 'http://{}/.well-known/challenges/{}/'.format(domain, challenge)
        headers = self._get_headers(domain)
        try:
            logger.info("Attempting to remove ACMEChallenge from server: %s" % challenge)
            requests.delete(url, headers=headers)
            logger.info("Successfully removed ACMEChallenge from server: %s" % challenge)
        except requests.RequestException as e:
            logger.warning("Encountered error while removing ACMEChallenge from server: %s" % challenge)


    def _get_headers(self, domain):
        private_key = self._get_private_key(domain)
        headers = {
            'Authorization': create_auth_header(username=self._get_username(), key=private_key)
        }
        return headers


    def _get_private_key(self, domain):
        filename = 'certbot_django_id_rsa_{}'.format(domain).replace('.', '')
        private_key_file = os.path.join(self._get_key_dir(), filename)
        private_key = None

        try:
            with open(private_key_file, 'r') as keyfile:
                private_key = keyfile.read()
        except (IOError, OSError):
            if self.config.noninteractive_mode:
                raise errors.PluginError('Could not read file from %s' % self._get_key_dir())

        if not private_key:
            private_key, public_key = generate_key_pair()
            try:
                logger.info('Trying to load private key: %s' % private_key_file)
                with open(private_key_file, 'w') as keyfile:
                    keyfile.write(private_key)
            except (IOError, OSError):
                raise errors.PluginError('Could not write private key to %s' % self._get_key_dir())
            msg = (
                "Couldn't read private key from {private_key_file}, so a new key has been generated and saved. "
                "Please go your website ({url}) and add the following public key to the user '{username}'. "
                "Please also make sure that '{username}' is a staff user and has permission to add and "
                "delete AcmeChallenge objects."
                "\n\n"
                "{public_key}"
            )
            msg = msg.format(
                private_key_file=private_key_file,
                url="http://{}/admin/asymmetric_jwt_auth/publickey/add/".format(domain),
                username=self._get_username(),
                public_key=public_key)
            display = zope.component.getUtility(interfaces.IDisplay)
            display.notification(msg, wrap=False, force_interactive=True)

        return private_key


    def _get_username(self):
        return self._get_config('username', 'Django username', _validate_username)


    def _get_key_dir(self):
        return self._get_config('key-directory', 'key storage directory', _validate_key_dir)


    def _get_config(self, option, name, validation_fn):
        value = self.conf(option)
        if not value:
            value = self._prompt_for_config("Input the {}".format(name), validation_fn)
            setattr(self.config, self.dest(option), value)
        if not value:
            cli_flag = '--{0}'.format(self.option_name(option))
            raise errors.PluginError('Must provide {} with flag {}'.format(name, cli_flag))
        return validation_fn(value)


    def _prompt_for_config(self, prompt, validation_fn=lambda value: value):
        display = zope.component.getUtility(interfaces.IDisplay)
        while True:
            code, value = display.directory_select(prompt, force_interactive=True)
            if code == display_util.HELP:
                return None
            elif code == display_util.CANCEL:
                return None
            else:
                try:
                    return validation_fn(value)
                except errors.PluginError as error:
                    display.notification(str(error), pause=False)
